@@ -5,6 +5,8 @@ from os import listdir, mkdir, walk
 from os.path import abspath, dirname, isdir, isfile, join
 from sys import version_info
 
+from .utils.node_lambda import parallel_node_eval
+
 try:
     from tomllib import loads as toml_load
 except ImportError:
@@ -47,6 +49,7 @@ DIRNAME_HOOKS = "hooks"
 DIRNAME_ITEM_TYPES = "items"
 DIRNAME_LIBS = "libs"
 FILENAME_GROUPS = "groups.py"
+FILENAME_MAGIC_STRINGS = "magic_strings.py"
 FILENAME_NODES = "nodes.py"
 FILENAME_REQUIREMENTS = "requirements.txt"
 
@@ -268,6 +271,7 @@ class Repository(MetadataGenerator):
         self.group_dict = {}
         self.node_dict = {}
         self.node_attribute_functions = {}
+        self.magic_string_functions = {}
         self._get_all_attr_code_cache = {}
         self._get_all_attr_result_cache = {}
 
@@ -309,7 +313,6 @@ class Repository(MetadataGenerator):
         if group.name in names(self.groups):
             raise RepositoryError(_("you cannot have two groups "
                                     "both named '{}'").format(group.name))
-        group.repo = self
         self.group_dict[group.name] = group
 
     def add_node(self, node):
@@ -322,8 +325,6 @@ class Repository(MetadataGenerator):
         if node.name in names(self.nodes):
             raise RepositoryError(_("you cannot have two nodes "
                                     "both named '{}'").format(node.name))
-
-        node.repo = self
         self.node_dict[node.name] = node
 
     @cached_property
@@ -468,6 +469,26 @@ class Repository(MetadataGenerator):
                 result[entity_name] = infodict
         return result
 
+    def get_magic_strings(self):
+        if not isfile(self.magic_strings_file):
+            return
+
+        def magic_string(func):
+            self.magic_string_functions[func.__name__] = func
+            return func
+
+        # We do not store the gotten attrs anywhere, because we're
+        # only interested in the defined magic strings.
+        self.get_all_attrs_from_file(
+            self.magic_strings_file,
+            base_env={
+                'libs': self.libs,
+                'magic_string': magic_string,
+                'repo_path': self.path,
+                'vault': self.vault,
+            },
+        )
+
     def items_from_dir(self, path):
         """
         Looks for Item subclasses in the given path.
@@ -536,6 +557,9 @@ class Repository(MetadataGenerator):
         """
         Returns a list of nodes where every node is a member of every
         group given.
+
+        :param group_names: list of names of the groups to check for
+        :return list of nodes where every node is a member of every group given.
         """
         base_group = set(self.get_group(group_names[0]).nodes)
         for group_name in group_names[1:]:
@@ -551,6 +575,9 @@ class Repository(MetadataGenerator):
         """
         Returns all nodes that are a member of at least one of the given
         groups.
+
+        :param group_names: list of names of the groups to check for
+        :return list of nodes which are in at least one of given groups
         """
         for node in self.nodes:
             if node.in_any_group(group_names):
@@ -559,9 +586,125 @@ class Repository(MetadataGenerator):
     def nodes_in_group(self, group_name):
         """
         Returns a list of nodes in the given group.
+
+        :param group_name: name of the group to check for
+        :return list of nodes which are in the given group
         """
         return self.nodes_in_all_groups([group_name])
 
+    def nodes_not_in_group(self, group_name):
+        """
+        Returns a list of nodes not in the given group.
+
+        :param group_name: name of the group to check for
+        :return list of nodes which are not in the given group
+        """
+        return [
+            node
+            for node in self.nodes
+            if not node.in_group(group_name)
+        ]
+
+    def nodes_with_bundle(self, bundle_name):
+        """
+        Returns a list of nodes that do have the given bundle.
+
+        :param bundle_name: name of the bundle to check for
+        :return list of nodes which have the given bundle
+        """
+        return [
+            node
+            for node in self.nodes
+            if bundle_name in names(node.bundles)
+        ]
+
+    def nodes_without_bundle(self, bundle_name):
+        """
+        Returns a list of nodes that do not have the given bundle.
+
+        :param bundle_name: name of the bundle to check for
+        :return list of nodes which do not have the given bundle
+        """
+        return [
+            node
+            for node in self.nodes
+            if bundle_name not in names(node.bundles)
+        ]
+
+    def nodes_matching_lambda(self, lambda_str, lambda_workers=None):
+        """
+        Returns a list of nodes matching the lambda.
+
+        Example:
+            nodes = repo.nodes_matching_lambda("lambda:node.metadata_get('foo/magic', 47) < 3")
+
+        :param lambda_str: string to evaluate as python code with `node` being one of the nodes,
+            expected to return a value that can be interpreted as boolean
+        :param lambda_workers: number of parallel workers used to check lambda condition on every node
+        :return list of nodes matching the given lambda
+        """
+        result_items = parallel_node_eval(
+            self.nodes,
+            lambda_str,
+            lambda_workers,
+        ).items()
+
+        return [
+            self.get_node(node_name)
+            for node_name, result in result_items
+            if result
+        ]
+
+    def nodes_matching(self, target_strings, lambda_workers=None):
+        """
+        Returns a list of nodes matching any of the given target-strings. This is the same API that is used by
+        all the bw commandlines, i.e. `bw items` or `bw apply` to select which nodes to operate on.
+
+        Example:
+            nodes = repo.nodes_matching(['wi-5.s2s', 'wi-5.routing'])
+
+        :param target_strings: expression to select target nodes:
+        my_node            # to select a single node
+        my_group           # all nodes in this group
+        bundle:my_bundle   # all nodes with this bundle
+        !bundle:my_bundle  # all nodes without this bundle
+        !group:my_group    # all nodes not in this group
+        "lambda:node.metadata_get('foo/magic', 47) < 3"
+        # all nodes whose metadata["foo"]["magic"] is less than three
+
+        :param lambda_workers: number of parallel workers to check lambda condition on nodes
+        :return list of nodes matching any of the given target-strings
+        """
+        if isinstance(target_strings, str):
+            target_strings = [target_strings]
+
+        targets = set()
+        for name in target_strings:
+            name = name.strip()
+            if name.startswith("bundle:"):
+                bundle_name = name.split(":", 1)[1]
+                targets.update(self.nodes_with_bundle(bundle_name))
+            elif name.startswith("!bundle:"):
+                bundle_name = name.split(":", 1)[1]
+                targets.update(self.nodes_without_bundle(bundle_name))
+            elif name.startswith("!group:"):
+                group_name = name.split(":", 1)[1]
+                targets.update(self.nodes_in_group(group_name))
+            elif name.startswith("lambda:"):
+                lambda_str = name.split(":", 1)[1]
+                targets.update(self.nodes_matching_lambda(lambda_str, lambda_workers))
+            else:
+                try:
+                    targets.add(self.get_node(name))
+                except NoSuchNode:
+                    try:
+                        group = self.get_group(name)
+                        targets.update(group.nodes)
+                    except NoSuchGroup:
+                        continue
+
+        return list(targets)
+
     def metadata_hash(self):
         repo_dict = {}
         for node in self.nodes:
@@ -604,11 +747,14 @@ class Repository(MetadataGenerator):
             if validate_name(dir_entry):
                 self.bundle_names.append(dir_entry)
 
+        # populate magic strings
+        self.get_magic_strings()
+
         # populate groups
         toml_groups = self.nodes_or_groups_from_dir("groups")
         self.group_dict = {}
-        for group in self.nodes_or_groups_from_file(self.groups_file, 'groups', toml_groups):
-            self.add_group(Group(*group))
+        for group_name, group_attrs in self.nodes_or_groups_from_file(self.groups_file, 'groups', toml_groups):
+            self.add_group(Group(group_name, attributes=group_attrs, repo=self))
 
         # populate items
         self.item_classes = list(self.items_from_dir(items.__path__[0]))
@@ -618,8 +764,8 @@ class Repository(MetadataGenerator):
         # populate nodes
         toml_nodes = self.nodes_or_groups_from_dir("nodes")
         self.node_dict = {}
-        for node in self.nodes_or_groups_from_file(self.nodes_file, 'nodes', toml_nodes):
-            self.add_node(Node(*node))
+        for node_name, node_attrs in self.nodes_or_groups_from_file(self.nodes_file, 'nodes', toml_nodes):
+            self.add_node(Node(node_name, attributes=node_attrs, repo=self))
 
     @cached_property
     def revision(self):
@@ -633,6 +779,7 @@ class Repository(MetadataGenerator):
         self.items_dir = join(self.path, DIRNAME_ITEM_TYPES)
         self.groups_file = join(self.path, FILENAME_GROUPS)
         self.libs_dir = join(self.path, DIRNAME_LIBS)
+        self.magic_strings_file = join(self.path, FILENAME_MAGIC_STRINGS)
         self.nodes_file = join(self.path, FILENAME_NODES)
 
         self.hooks = HooksProxy(self, self.hooks_dir)
